#include <algorithm>
#include "CGAL/Random.h"
#include <CGAL/bounding_box.h>
#include "./dnet.h"


namespace dnet{

  double bg_c1 = 30.0; //size of global epsnet is with eps=bg_c1/OPT.
  double bg_c2 = 12.0; //size of phase_epsnet is with eps=bg_c2/OPT.
  double bg_c3 = 2.0;  //reweigh a disk only if weight < bg_c3/OPT
  double bg_c4 = 2.0;  //reweighing factor
  double bg_c5 = 0.6;  //finish algorithm if remaining_hittingset.size < bg_c5*OPT.


//#define STATISTICS_ENABLED
#ifdef STATISTICS_ENABLED
  size_t nR,nR_opt,nS,nS_opt,nQ,nQ_opt,nP,nP_opt,nRR,nRR_opt,nD,nD_opt; // only for some stats for the paper
# define STATS(x) x
#else
# define STATS(x) (void) 0
#endif

  class ComparePoints{
  public:
    ComparePoints(const Point& p):origin(p){}
    bool operator() (const Point & p, const Point & q){
      int quad_p, quad_q;

      if ( p.x() - origin.x() >= 0 )
        quad_p = ( p.y() - origin.y() >= 0 ) ? 1 : 4 ;
      else
        quad_p = ( p.y() - origin.y() >= 0 ) ? 2 : 3 ;

      if ( q.x() - origin.x() >= 0 )
        quad_q = ( q.y() - origin.y() >= 0 ) ? 1 : 4 ;
      else
        quad_q = ( q.y() - origin.y() >= 0 ) ? 2 : 3 ;

      if (quad_p == quad_q) 
      {
        if ( orientation(origin, p, q) > 0 )
          return 1;
        else 
          return 0;
      }

      return ( quad_p < quad_q ? 1 : 0 );
    }
  private:
    const Point origin;
  };

  class CompareFaces{
  public:
    CompareFaces(DelaunayInfo & dt,const Point & pt):dt(dt),origin(pt){}
    bool operator() (const Face_handle & p_face, const Face_handle & q_face)
    {
      Point p = dt.dual(p_face);
      Point q = dt.dual(q_face);
      ComparePoints cmpfp(origin);
      return cmpfp(p, q);
    }
  private:
    DelaunayInfo & dt;
    const Point origin;
  };


  void preprocess( const std::vector<Point> & input_points, const std::vector<double> & weights, const std::vector<Circle> & input_disks, 
    std::vector<size_t> & disks_remaining, std::vector<size_t> & epsnet , DelaunayInfo & bg_dt, std::vector< std::vector<Face_handle> >& bg_sortedFaces )
  {
    //assumes epsnet is a sorted list of indices.
    const size_t t = epsnet.size();
    std::vector< std::pair<Point,unsigned> > inputPoints_w_info;

    for (size_t i=0;i<t;++i)
      inputPoints_w_info.push_back( std::make_pair( input_points[epsnet[i]], (unsigned) i) );

    bg_dt.clear();
    bg_dt.insert(inputPoints_w_info.begin(), inputPoints_w_info.end());

    Face_handle face;
    size_t epsnet_it = 0;
    std::vector<Face_handle> containingDisks;
    for (size_t i=0; i < input_points.size(); ++i)
    {
      if (epsnet[epsnet_it] == i) {epsnet_it++; continue;}

      face = bg_dt.locate(input_points[i],face);
      containingDisks.clear();
      bg_dt.get_conflicts(input_points[i], std::back_inserter(containingDisks), face); 

      for (size_t j = 0; j < containingDisks.size() ; ++j)
        containingDisks[j]->info().diskpoints.push_back(i);
    }

    for (Face_handle fit = bg_dt.faces_begin(); fit != (Face_handle) bg_dt.faces_end(); fit++)
      std::sort( fit->info().diskpoints.begin(), fit->info().diskpoints.end() );

    bg_sortedFaces.clear();
    bg_sortedFaces.resize( t );

    Face_circulator v_face, f_tmp;
    for(Vertex_handle vit = bg_dt.finite_vertices_begin(); vit != (Vertex_handle) bg_dt.finite_vertices_end(); ++vit)
    {
      bg_sortedFaces[vit->info()].clear();
      v_face = bg_dt.incident_faces(vit);
      f_tmp = v_face;
      do {
        if ( !bg_dt.is_infinite(f_tmp) ) 
          bg_sortedFaces[vit->info()].push_back( (Face_handle) f_tmp );
        f_tmp++;
      } while (f_tmp != v_face);
    }

    for (size_t i = 0; i<t; ++i) {
      CompareFaces cmpf(bg_dt,input_points[epsnet[i]]);
      std::sort( bg_sortedFaces[i].begin(), bg_sortedFaces[i].end(), cmpf);
    }

    disks_remaining.clear();
    for (size_t i=0;i<input_disks.size();++i) 
    {
      Circle D = input_disks[i];
      Point q = bg_dt.nearest_vertex( D.center() )->point();

      if ( D.squared_radius() < squared_distance(q, D.center()) )
        disks_remaining.push_back(i);
    }
  }


  double computePointsinDisk(const std::vector<Point> & input_points, const std::vector<double> & weights, 
    const std::vector<Circle> & input_disks, std::vector<size_t> & disks_remaining,  
    std::vector<size_t> & epsnet,  const Circle & D, std::vector<size_t> & output_list , DelaunayInfo & bg_dt, std::vector< std::vector<Face_handle> > & bg_sortedFaces )
  {
    Vertex_handle vit = bg_dt.nearest_vertex( D.center() );
    Point o = input_points[epsnet[vit->info()]];

    //binary search in the array sortedFaces[vit->info()] for D.center()
    size_t start = 0, end = bg_sortedFaces[vit->info()].size()-1, mid;
    Face_handle left, right;

    ComparePoints cmpfp(o);
    if ( !cmpfp( bg_dt.dual(bg_sortedFaces[vit->info()][start]), D.center()) || cmpfp( bg_dt.dual(bg_sortedFaces[vit->info()][end]), D.center()) )
    {
      start = 0;
      end = bg_sortedFaces[vit->info()].size()-1;
    }
    else {
      while (start != end-1)
      {
        mid = (start+end)/2;
        if ( cmpfp( bg_dt.dual(bg_sortedFaces[vit->info()][mid]), D.center()) ) 
          start = mid;
        else end = mid;
      }
    }
    left = bg_sortedFaces[vit->info()][start];
    right = bg_sortedFaces[vit->info()][end];


    double weight = 0.0;
    size_t left_i = 0, right_i = 0;
    output_list.clear();
    while (right_i < right->info().diskpoints.size() || left_i < left->info().diskpoints.size())
    {
      if ( right_i == right->info().diskpoints.size() || 
        (left_i < left->info().diskpoints.size() && left->info().diskpoints[left_i] <= right->info().diskpoints[right_i]) )
      {
        if ( D.squared_radius() >= squared_distance( input_points[left->info().diskpoints[left_i]], D.center() )  ) {
          output_list.push_back(left->info().diskpoints[left_i]);
          weight +=  weights[left->info().diskpoints[left_i]];
        }
        if (right_i < right->info().diskpoints.size() && 
          left->info().diskpoints[left_i] == right->info().diskpoints[right_i]) right_i++;
        left_i++;
      }
      else //if (left->info().diskpoints[left_i] > right->info().diskpoints[right_i])
      {
        if ( D.squared_radius() >= squared_distance( input_points[right->info().diskpoints[right_i]], D.center() )  ) {
          output_list.push_back(right->info().diskpoints[right_i]);
          weight +=  weights[right->info().diskpoints[right_i]];
        }
        right_i++;
      }
    }

    return weight;
  }

  size_t totalRangeReport;
  size_t totalDoubling;

  int runphase( const std::vector<Point> & input_points, std::vector<double> & weights, 
    const std::vector<Circle> & input_disks, std::vector<size_t> & disks_remaining, std::vector<size_t> & epsnet, std::vector<size_t> & phaseepsnet_union_S, double phase_epsilon, 
    double threshold, double reweigh_constant, size_t OPT , DelaunayInfo& bg_dt, std::vector< std::vector<Face_handle> >& bg_sortedFaces , double & bg_W)
  {
    size_t tmp;
    size_t num_doublings = 0;
    size_t num_range_report = 0;
    double weight;

    std::vector<size_t> points_toreweigh;
    points_toreweigh.reserve(input_points.size());

    Delaunay phase_epsnet_dt;
    size_t disks_hitbyphaseepsnet = 0;

    phaseepsnet_union_S.clear();
    enet::randomSampleEpsnet(input_points, weights, phase_epsilon, phaseepsnet_union_S, tmp);
    const size_t t = phaseepsnet_union_S.size();
    STATS(nR=t);

    std::cout << "Size of phase epsnet: " << t << std::endl;

    std::vector<Point> inputPoints;
    for (size_t i=0;i<t;++i) inputPoints.push_back( input_points[phaseepsnet_union_S[i]] );
    phase_epsnet_dt.insert(inputPoints.begin(), inputPoints.end());

    std::random_shuffle ( disks_remaining.begin(),  disks_remaining.end() ); 
    for (std::vector<size_t>::iterator lit =  disks_remaining.begin(); lit !=  disks_remaining.end(); ++lit)
    {
      Circle D = input_disks[*lit];
      Point q = phase_epsnet_dt.nearest_vertex( D.center() )->point();

      if ( D.squared_radius() < squared_distance(q, D.center()) )
      {
        num_range_report++;
        totalRangeReport++;
        weight = computePointsinDisk(input_points, weights, input_disks, disks_remaining, epsnet, D, points_toreweigh,bg_dt,bg_sortedFaces);

        //std::cout << "verfication for computePointsinDisk: " << verifyPointsinDisk(rs, epsnet, phase_epsnet, D, points_toreweigh) << std::endl;
        if (points_toreweigh.size() == 0) { std::cout << "ERROR, no points in Disk\n"; exit(0); }

        if (weight < bg_W*threshold) 
        {
          for (size_t i = 0;i<points_toreweigh.size();++i)
          {
            bg_W += reweigh_constant  *  weights[points_toreweigh[i]];
            weights[points_toreweigh[i]] *= reweigh_constant;
          }
          num_doublings++;
          totalDoubling++;
        }
        size_t random_point_to_add = rand() % points_toreweigh.size();
        phaseepsnet_union_S.push_back(points_toreweigh[random_point_to_add]); //add one point from each unhit disk to remaining_hittingset.
        phase_epsnet_dt.push_back( input_points[points_toreweigh[random_point_to_add]] ); //adding for future nearest search
      }
      else disks_hitbyphaseepsnet++;

      if (num_doublings > 2*OPT) {
        //std::cout << "Hit by phase epsnet: " << disks_hitbyphaseepsnet << ", number of reweighings: " << num_doublings << std::endl;
        std::cout << "MoreThanOPT: RangeReport: " << num_range_report << " Doubling: " << num_doublings << std::endl;
        return -1;
      }
    }

    std::cout << "Hit by phase epsnet: " << disks_hitbyphaseepsnet << std::endl;
    std::cout << "RangeReport: " << num_range_report << " Doubling: " << num_doublings << std::endl;
    std::cout << "Size of S: " << (int)( disks_remaining.size() - disks_hitbyphaseepsnet) << std::endl;
    return (int)( disks_remaining.size() - disks_hitbyphaseepsnet);
  }


  void bg_solve_with_opt( const std::vector<Point> & input_points, const std::vector<Circle> & input_disks, std::vector<size_t> & curr_hittingset, size_t OPT)
  {
    size_t tmp;

    std::vector<double> weights;
    weights.resize(input_points.size(), 1.0);

    std::vector<size_t> epsnet;
    std::vector<size_t> phaseepsnet_union_S;

    enet::randomSampleEpsnet(input_points, weights, bg_c1/(double)OPT, epsnet, tmp);
    epsnet.push_back(input_points.size()-4);
    epsnet.push_back(input_points.size()-3);
    epsnet.push_back(input_points.size()-2);
    epsnet.push_back(input_points.size()-1);
    std::sort(epsnet.begin(), epsnet.end());

    std::vector<size_t> disks_remaining;
    DelaunayInfo bg_dt;
    std::vector< std::vector<Face_handle> > bg_sortedFaces;
    preprocess(input_points, weights, input_disks, disks_remaining, epsnet,bg_dt,bg_sortedFaces);
    std::cout << "Size of global epsnet: " << epsnet.size() << std::endl;
    STATS(nQ = epsnet.size());
    std::cout << "Remaining disks: " <<  disks_remaining.size() << std::endl;

    curr_hittingset.clear();
    for (size_t i=0;i<input_points.size();++i) curr_hittingset.push_back(i);

    totalDoubling = totalRangeReport = 0;
    size_t max_iterations = (size_t) (log((double)input_points.size())+1);
    int S = 0;
    for (size_t i=0;i< max_iterations;i++) {
      double bg_W = 0.0;
      for (size_t j=0;j<input_points.size();++j)
        bg_W +=  weights[j];

      std::cout << "*RUNNING PHASE " << i << "*" << std::endl;
      S = runphase(input_points, weights, input_disks, disks_remaining, epsnet, phaseepsnet_union_S, bg_c2/(double)OPT, bg_c3/(double)OPT, bg_c4, OPT,bg_dt,bg_sortedFaces,bg_W);
      STATS(nS=S);
      //S contains the size of the set S of that phase (see paper).
      if ( S >= 0 && epsnet.size() + phaseepsnet_union_S.size()  < curr_hittingset.size()  )
      {
        curr_hittingset.clear();
        std::sort (phaseepsnet_union_S.begin(), phaseepsnet_union_S.end());
        std::set_union (epsnet.begin(), epsnet.end(), phaseepsnet_union_S.begin(), phaseepsnet_union_S.end(), std::back_inserter(curr_hittingset));	
      }
      if ( S >= 0 && S < bg_c5*OPT ) {
        STATS(nP=i+1);
        STATS(nRR=totalRangeReport);
        STATS(nD = totalDoubling);
        std::cout << "Number of iterations after which we found success is: " << i+1 << std::endl;
        std::cout << "Total rangeReport: " << totalRangeReport << " total doubling: " << totalDoubling << std::endl;
        return;
      }
    } 
    STATS(nP = max_iterations);
    STATS(nRR = totalRangeReport); 
    STATS(nD = totalDoubling);
    std::cout << "Number of iterations after which we found success is MAX: " << max_iterations << std::endl;
    std::cout << "Total rangeReport: " << totalRangeReport << " total doubling: " << totalDoubling << std::endl;
  }


  void bg_solve( std::vector<Point> & input_points, const std::vector<Circle> & input_disks, std::vector<size_t> & Hittingset, size_t OPT )
  {
    CGAL::spatial_sort(input_points.begin(), input_points.end());

    //add four points of the bounding box to the pointset.
    K::Iso_rectangle_2 bb = CGAL::bounding_box(input_points.begin(), input_points.end());
    double xlen = bb.xmax() - bb.xmin(), ylen =  bb.ymax() - bb.ymin();
    input_points.push_back(Point(bb.xmax()+2*xlen, bb.ymax()+2*ylen));
    input_points.push_back(Point(bb.xmax()+2*xlen, bb.ymin()-2*ylen));
    input_points.push_back(Point(bb.xmin()-2*xlen, bb.ymax()+2*ylen));
    input_points.push_back(Point(bb.xmin()-2*xlen, bb.ymin()-2*ylen));

    Hittingset.resize(input_points.size());

    if (OPT != 0) 
      bg_solve_with_opt(input_points, input_disks, Hittingset, OPT);
    else 
    {
      for (size_t i=0;i<input_points.size();++i) Hittingset[i] = i;
      std::vector<size_t> curr_hittingset;
      curr_hittingset.reserve(input_points.size());
      std::vector<int> opt_vals;
      for (int i=2;i<input_points.size(); i *= 2)
        opt_vals.push_back(i);
//#pragma omp parallel for firstprivate(curr_hittingset)
      for (int i=0;i<opt_vals.size(); ++i)
      {
        std::cout << "\n\n***RUNNING WITH OPT = " << opt_vals[i] << "***" << std::endl << std::endl;
        bg_solve_with_opt(input_points, input_disks, curr_hittingset, opt_vals[i]);
#pragma omp critical (updateBestHittingSet)
        {
          std::cout << "Size of hitting set in iteration " << opt_vals[i] << " is: " << curr_hittingset.size() << std::endl << std::endl;
          if (curr_hittingset.size() < Hittingset.size()) 
          {
            Hittingset = curr_hittingset;
            STATS(nR_opt = nR);
            STATS(nS_opt = nS);
            STATS(nQ_opt = nQ);
            STATS(nP_opt = nP);
            STATS(nRR_opt = nRR);
            STATS(nD_opt = nD);
          }
        }
      }
    }
    STATS(std::cout << "Q: "<<nQ_opt<<" R: "<<nR_opt<<" S: "<<nS_opt<<" nPhases: "<<nP_opt<<std::endl);
    STATS(std::cout<< "RangeReportings: "<<nRR_opt<<" Doubling: "<<nD_opt<<std::endl);
    //remove the four bounding box points from rs.points and Hittingset (which is assumed to be sorted).
    input_points.pop_back(); input_points.pop_back(); input_points.pop_back(); input_points.pop_back();
    while (Hittingset.back() >= input_points.size()) Hittingset.pop_back();
  }
}
